using System.Net;
using Autofac;
using FitConnect.Encryption;
using FitConnect.Models;
using FitConnect.Models.v1.Api;
using FitConnect.Services;
using FitConnect.Services.Interfaces;
using FitConnect.Services.Models.v1.Submission;
using IdentityModel;
using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using NJsonSchema;
using NJsonSchema.Validation;

namespace FitConnect;

public abstract class FitConnectClient {
    public const string GuidPattern =
        "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$";

    public const string ServiceIdentifierPattern =
        "^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$";

    public const string LeikaKeyPattern = @"^99\d{12}$";

    internal const string VersionHeaderPattern =
        @"^(FITConnectDotNetSDK)\/(((\d+\.){2,3}\d+(-[a-z]+\.\d+)?)|(local-build))\s{1}\((commit:)([a-zA-Z0-9]{5,}|(unreleased));\s{1}(os):(.*)\)$";

    private readonly List<string>? _privateDecryptionKeys;
    private readonly string? _privateKeySigning;
    private readonly string? _publicKeyEncryption;
    private readonly string? _publicKeySignatureVerification;
    protected readonly bool VerifiedCertificatesAreMandatory;

    protected FitConnectClient(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        ILogger? logger = null,
        List<string>? privateDecryptionKeys = null,
        string? privateKeySigning = null,
        string? publicKeyEncryption = null,
        string? publicKeySignatureVerification = null
    ) : this() {
        _privateDecryptionKeys = privateDecryptionKeys;
        _privateKeySigning = privateKeySigning;
        _publicKeyEncryption = publicKeyEncryption;
        _publicKeySignatureVerification = publicKeySignatureVerification;
        OAuthService = new OAuthService(environment.TokenUrl, "V1", clientId, clientSecret, logger);
        SubmissionService =
            new SubmissionService(environment.SubmissionUrl[0], OAuthService,
                publicKeySignatureVerification == null
                    ? null
                    : new JsonWebKey(publicKeySignatureVerification),
                logger: logger);
        RouteService = new RouteService(environment.RoutingUrl, logger: logger);
        CasesService = new CasesService(environment.SubmissionUrl[0], OAuthService, logger: logger);
        DestinationService =
            new DestinationService(environment.SubmissionUrl[0], OAuthService, logger: logger);
        Logger = logger;
        VerifiedCertificatesAreMandatory = environment.VerifiedCertificatesAreMandatory;
    }

    protected FitConnectClient(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        IContainer container,
        List<string>? privateDecryptionKeys = null,
        string? privateKeySigning = null,
        string? publicKeyEncryption = null,
        string? publicKeySignatureVerification = null
    ) : this() {
        _privateDecryptionKeys = privateDecryptionKeys;
        _privateKeySigning = privateKeySigning;
        _publicKeyEncryption = publicKeyEncryption;
        _publicKeySignatureVerification = publicKeySignatureVerification;
        OAuthService = container.Resolve<IOAuthService>();
        SubmissionService = container.Resolve<ISubmissionService>();
        RouteService = container.Resolve<IRouteService>();
        CasesService = container.Resolve<ICasesService>();
        DestinationService = container.Resolve<IDestinationService>();
        Logger = container.Resolve<ILogger>();
    }


    private FitConnectClient() {
        // Take care of OAuthService if make the constructor public
        Encryption = new FitEncryption(Logger);
    }


    public IOAuthService OAuthService { get; } =
        null!; // To resolve the warning, is only applicable for the **private** constructor

    protected ISubmissionService SubmissionService { get; } = null!;
    protected IRouteService RouteService { get; } = null!;
    protected IDestinationService DestinationService { get; } = null!;
    protected ICasesService CasesService { get; } = null!;
    protected ILogger? Logger { get; }
    protected FitEncryption Encryption { get; set; }

    /// <summary>
    ///     Validates <paramref name="dataString" /> against <paramref name="schemaString" /> and returns
    ///     the validation errors if any appear
    /// </summary>
    public static async Task<ICollection<ValidationError>> ValidateSchemaAsync(
        string dataString, string schemaString) {
        var schema = await JsonSchema.FromJsonAsync(schemaString);
        return schema.Validate(dataString);
    }

    public void SetProxy(WebProxy proxy) {
        OAuthService.Proxy = proxy;
        SubmissionService.Proxy = proxy;
        DestinationService.Proxy = proxy;
        RouteService.Proxy = proxy;
    }

    public WebProxy? GetProxy() {
        return OAuthService.Proxy;
    }

    public static bool ValidateCallback(string hmac, string callbackSecret,
        long timestamp, string httpBody) {
        if (timestamp < DateTime.Now.AddMinutes(ProjectSpecification.MaxCallbackAge * -1)
                .ToEpochTime())
            throw new ArgumentException("Request is too old");
        var hmacCalculation = EncryptionSettings.CalculateCallbackHmac(callbackSecret, timestamp, httpBody);
        return hmac ==  Convert.ToHexString(hmacCalculation).ToLower();
    }


    public static bool ValidateCallback(string callbackSecret, HttpRequest request) {
        if (!request.Headers.ContainsKey(ProjectSpecification.CallbackTimestamp))
            throw new ArgumentException("Missing callback-timestamp header");

        var timeStampString = request.Headers[ProjectSpecification.CallbackTimestamp].ToString();
        if (!long.TryParse(timeStampString, out var timestamp))
            throw new ArgumentException("Invalid callback-timestamp header");


        var authentication = request.Headers[ProjectSpecification.CallbackAuthentication];
        using var requestStream = request.Body;
        var content = new StreamReader(requestStream).ReadToEnd().Trim();

        var result = ValidateCallback(authentication, callbackSecret, timestamp, content);
        if (!result)
            throw new ArgumentException("Verified request does not match authentication");
        return true;
    }


    /// <summary>
    ///     Retrieve the events for the submission
    /// </summary>
    /// <param name="caseId"></param>
    /// <param name="destinationId"></param>
    /// <param name="skipTest"></param>
    /// <returns></returns>
    public virtual async Task<List<SecurityEventToken>> GetEventLogAsync(Guid caseId,
        Guid destinationId,
        bool skipTest = false) {
        var events =
            (await SubmissionService.GetEventLogAsync(caseId, destinationId, skipTest))
            ?.Select(SecurityEventToken.FromJwtEncodedString).ToList() ??
            new List<SecurityEventToken>();
        return events;
    }

    /// <summary>
    ///     Retrieve the events for the submission
    /// </summary>
    /// <param name="submission"></param>
    /// <param name="skipTest"></param>
    /// <returns></returns>
    public virtual async Task<List<SecurityEventToken>> GetEventLogAsync(
        SubmissionForPickup submission,
        bool skipTest = false) {
        var events = await SubmissionService
            .GetEventLogAsync(submission.CaseId, submission.DestinationId, skipTest);
        return events?.Select(SecurityEventToken.FromJwtEncodedString).ToList() ??
               new List<SecurityEventToken>();
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    public async Task<SubmissionStatus> GetStatusForSubmissionAsync(
        SentSubmission submission) {
        var eventLog = await GetEventLogAsync(submission.CaseId, submission.DestinationId);

        var status = eventLog.Last(e => e.SubmissionId == submission.SubmissionId);
        List<Problems>? problems = null;

        if (status.Events?.AcceptSubmissionEvent != null) {
            if (!(status.Events.AcceptSubmissionEvent.AuthenticationTags ??
                  throw new InvalidOperationException()).AreEqualTo(submission
                    .AuthenticationTags)) {
                throw new AuthenticationTagException("Authentication tags are not equal");
            }

            problems = status.Events.AcceptSubmissionEvent.Problems;
        }
        else if (status.Events?.RejectSubmissionEvent != null)
            problems = status.Events.RejectSubmissionEvent.Problems;

        return new SubmissionStatus(status.EventType, problems);
    }

    protected virtual async Task<ICollection<ValidationError>> CheckDataSchemaAsync(Submission submission) {
        var dataSchemaObject = await JsonSchema.FromUrlAsync(submission.Metadata?.ContentStructure.Data.SubmissionSchema.SchemaUri);
        var jSchema = await NJsonSchema.JsonSchema.FromUrlAsync(submission.Metadata?.ContentStructure.Data.SubmissionSchema.SchemaUri);
        var isValidSchemaData = jSchema.Validate(submission.Data);

        return isValidSchemaData;
    }
    
    /// <summary>
    ///     Encrypt attachments (Anhänge)
    /// </summary>
    /// <param name="publicKey"></param>
    /// <param name="attachment"></param>
    /// <returns></returns>
    public KeyValuePair<Guid, string> Encrypt(string publicKey, Attachment attachment) {
        var content = Encryption.Encrypt(attachment.GetBytes(), publicKey);
        return new KeyValuePair<Guid, string>(attachment.Id, content);
    }

    /// <summary>
    ///     Encrypt attachments (Anhänge)
    /// </summary>
    /// <param name="publicKey">Public key for encryption</param>
    /// <param name="attachments">List of attachments to encrypt</param>
    /// <returns>List with ID and authentication tag</returns>
    public Dictionary<Guid, string> Encrypt(string publicKey,
        IEnumerable<Attachment> attachments) {
        return attachments.Select(a => Encrypt(publicKey, a))
            .ToDictionary(p => p.Key, p => p.Value);
    }
}

public static class FitConnectClientExtensions {
    /// <summary>
    /// </summary>
    /// <param name="host"></param>
    /// <param name="port"></param>
    /// <param name="username"></param>
    /// <param name="password"></param>
    /// <returns></returns>
    public static T WithProxy<T>(this T caller, string host, int port, string? username = null,
        string? password = null) where T : FitConnectClient {
        var proxy = new WebProxy(host, port);
        if (username != null && password != null)
            proxy.Credentials = new NetworkCredential(username, password);

        caller.SetProxy(proxy);
        return caller;
    }
}
