using System.Security;
using System.Text.RegularExpressions;
using Autofac;
using FitConnect.Encryption;
using FitConnect.Interfaces;
using FitConnect.Models;
using FitConnect.Models.Api.Metadata;
using FitConnect.Models.Api.Set;
using FitConnect.Services.Models.v1.Submission;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Attachment = FitConnect.Models.Attachment;
using Metadata = FitConnect.Models.Metadata;

namespace FitConnect;

/// <summary>
///     The fluent implementation for the <see cref="Sender" /> to encapsulate the FitConnect API.
/// </summary>
/// <example>
///     Reference for FluentSender
///     <code>
/// client.Sender
///     .Authenticate(clientId!, clientSecret!)
///     .CreateSubmission(new Submission { Attachments = new List() })
///     .UploadAttachments()
///     .WithData(data)
///     .Submit();
/// </code>
/// </example>
public class Sender : FitConnectClient, ISender {
    internal Sender(FitConnectEnvironment environment, string clientId, string clientSecret,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger) {
    }

    public Sender(FitConnectEnvironment environment, string clientId, string clientSecret,
        IContainer container) : base(environment, clientId, clientSecret,
        container) {
    }

    public string? PublicKey { get; set; }

    public async Task<string> GetPublicKeyForDestinationAsync(Guid destinationId) {
        var publicKey = await DestinationService.GetPublicKey(destinationId);

        var keyIsValid = new CertificateHelper(Logger).ValidateCertificateJsonWebKeyString(
            publicKey,
            VerifiedCertificatesAreMandatory ? LogLevel.Error : LogLevel.Warning);
        if (VerifiedCertificatesAreMandatory && !await keyIsValid)
            throw new SecurityException("Public key is not trusted");

        return publicKey;
    }


    public async Task<SentSubmission> SendAsync(SendableSubmission submission) {
        SubmissionSenderGuards(submission);
        var sendable = await CreateSubmission(submission);
        sendable.AddServiceType(submission.ServiceName!, submission.ServiceIdentifier!);

        sendable.Data = submission.Data;
        sendable.SchemaUri = submission.SchemaUri;

        sendable!.Attachments = new List<Attachment>();
        if (submission.Attachments != null)
            foreach (var attachment in submission.Attachments)
                sendable!.Attachments.Add(attachment);

        var created =
            await SubmissionService.CreateSubmissionAsync((CreateSubmissionDto)sendable);
        sendable.Id = created.SubmissionId;
        sendable.CaseId = created.CaseId;

        Logger?.LogInformation("Submission Id {CreatedSubmissionId}, CaseId {SubmissionCaseId}",
            created.SubmissionId, sendable.CaseId);

        var encryptedAttachments = Encrypt(PublicKey!, sendable.Attachments);
        await UploadAttachmentsAsync(sendable.Id!, encryptedAttachments);

        foreach (var (key, value) in encryptedAttachments) {
            var item = submission.Attachments!.Single(attachment => attachment.Id == key);
            item.AttachmentAuthentication = value.Split('.').Last();
        }

        return (SentSubmission)await Submit(sendable);
    }

    public override Task<List<SecurityEventToken>> GetEventLogAsync(Guid caseId,
        Guid destinationId, bool skipTest = false) {
        return base.GetEventLogAsync(caseId, destinationId, true);
    }

    public override Task<List<SecurityEventToken>> GetEventLogAsync(
        SubmissionForPickup submission, bool skipTest = false) {
        return base.GetEventLogAsync(submission, true);
    }

    /// <summary>
    ///     Sender for pre-encrypted data
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    /// <exception cref="ArgumentNullException"></exception>
    /// <exception cref="ArgumentException"></exception>
    public async Task<SentSubmission> SendAsync(SendableEncryptedSubmission submission) {
        await SubmissionSenderGuards(submission);

        var sendable = await CreateSubmission(submission);
        sendable.AddServiceType(submission.ServiceName!, submission.ServiceIdentifier!);
        sendable.EncryptedData = submission.Data;
        sendable.EncryptedMetadata = submission.Metadata;

        var announcedSubmission = (CreateSubmissionDto)sendable;
        announcedSubmission.AnnouncedAttachments = submission.Attachments?.Keys.ToList();

        var created =
            await SubmissionService.CreateSubmissionAsync(announcedSubmission);
        sendable.Id = created.SubmissionId;
        sendable.CaseId = created.CaseId;

        Logger?.LogInformation("Submission Id {CreatedSubmissionId}, CaseId {SubmissionCaseId}",
            created.SubmissionId, sendable.CaseId);

        Logger?.LogInformation("Uploading pre encrypted attachments");
        await UploadAttachmentsAsync(sendable.Id!, submission.Attachments!);

        if (submission.Attachments != null) {
            sendable.Attachments ??= new List<Attachment>();
            foreach (var (key, value) in submission.Attachments) {
                sendable.Attachments.Add(
                    Attachment.FromPreEncryptedAttachment(key, value.Split('.').Last()));
            }
        }
        return (SentSubmission)(await Submit(sendable));
    }

    private async Task<Submission> Submit(Submission submission) {
        if (submission == null) {
            Logger?.LogCritical("Submission is null on submit");
            throw new InvalidOperationException("Submission is not ready");
        }

        if (submission.EncryptedMetadata == null) {
            var metadata = CreateMetadata(submission);
            submission.GeneratedMetadata = metadata;
            Logger?.LogTrace("Metadata: {Metadata}", metadata);
            var valid = await JsonHelper.VerifyMetadata(metadata, null, Logger);
            if (!valid) {
                Logger?.LogError("Sending submission aborted due to validation errors");
                throw new InvalidOperationException("Submission is not ready");
            }

            Logger?.LogInformation("Metadata validation check, done");
            Logger?.LogInformation("Sending submission");
            var encryptedMeta = Encryption.Encrypt(metadata);
            Logger?.LogTrace("Encrypted metadata: {EncryptedMeta}", encryptedMeta);
            submission.EncryptedMetadata = encryptedMeta;
        }

        if (submission.EncryptedData == null)
            if (submission.Data != null)
                submission.EncryptedData = Encryption.Encrypt(submission.Data);

        var result = await SubmissionService
            .SubmitSubmission(submission.Id!, (SubmitSubmissionDto)submission);

        Logger?.LogInformation("Submission sent");
        return submission;
    }

    /// <summary>
    ///     Creates a new <see cref="SendableSubmission" /> on the FitConnect server.
    /// </summary>
    /// <param name="sendableSubmission"></param>
    /// <returns></returns>
    /// <exception cref="InvalidOperationException">If sender is not authenticated</exception>
    /// <exception cref="ArgumentException">If submission is not ready to be sent</exception>
    private async Task<Submission> CreateSubmission(SendableSubmission sendableSubmission) {
        Guid destinationId = sendableSubmission.DestinationId;
        PublicKey = await GetPublicKeyForDestinationAsync(destinationId);
        Encryption = new FitEncryption(Logger) { PublicEncryptionKey = PublicKey };
        return sendableSubmission.ToSubmission();
    }

    /// <summary>
    ///     Creates a new <see cref="SendableSubmission" /> on the FitConnect server.
    /// </summary>
    /// <param name="sendableSubmission"></param>
    /// <returns></returns>
    /// <exception cref="InvalidOperationException">If sender is not authenticated</exception>
    /// <exception cref="ArgumentException">If submission is not ready to be sent</exception>
    private async Task<Submission>
        CreateSubmission(ISendableSubmission sendableSubmission) {
        var destinationId = sendableSubmission.DestinationId;
        PublicKey = await GetPublicKeyForDestinationAsync(destinationId);
        Encryption = new FitEncryption(Logger) { PublicEncryptionKey = PublicKey };
        return new Submission {
            DestinationId = destinationId
        };
    }

    /// <summary>
    ///     Create Metadata incl. Hash
    /// </summary>
    /// <param name="submission"></param>
    /// <param name="dataHash"></param>
    /// <returns></returns>
    internal static string CreateMetadata(Submission submission,
        DataHash? dataHash = null) {
        var data = new Data {
            Hash = dataHash ?? FitEncryption.CalculateHash(submission.Data!),

            SubmissionSchema = new Fachdatenschema {
                SchemaUri = submission.ServiceType.Identifier != null
                    ? new Uri(submission.ServiceType.Identifier)
                    : null,
                MimeType = submission.DataMimeType
            }
        };
        var contentStructure = new ContentStructure {
            Data = data,
            Attachments = submission.Attachments?.Select(attachment =>
                new Models.Api.Metadata.AttachmentMetadata {
                    Description = attachment.Description,
                    AttachmentId = attachment.Id,
                    MimeType = attachment.MimeType,
                    Filename = attachment.Filename,
                    Purpose = attachment.Purpose,
                    Hash = attachment.Hash
                }).ToList() ?? new List<Models.Api.Metadata.AttachmentMetadata>()
        };

        var metadata = new Metadata {
            SchemaUri = new Uri(Metadata.SchemaUrl),
            ContentStructure = contentStructure,
            PublicServiceType = new Verwaltungsleistung {
                Identifier = submission.ServiceType.Identifier,
                Name = submission.ServiceType.Name
            },
            ReplyChannel = submission.ReplyChannel,
            PaymentInformation = submission.PaymentInformation,
            AuthenticationInformation = submission.AuthenticationInformation?.ToList()
        };
        return JsonConvert.SerializeObject(metadata, Formatting.None, new JsonSerializerSettings() {
            Converters = new List<JsonConverter>() {
                new StatusConverter(),
                new PaymentMethodConverter(),
                new AuthenticationInformationTypeConverter()
            },
            NullValueHandling = NullValueHandling.Ignore,
        });
    }

    /// <summary>
    ///     Finding Areas
    /// </summary>
    /// <param name="filter"></param>
    /// <param name="offset"></param>
    /// <param name="limit"></param>
    /// <returns></returns>
    public async Task<IEnumerable<Area>> GetAreas(string filter, int offset = 0,
        int limit = 100) {
        var dto = await RouteService.FindAreas(new List<string>() { filter }, offset, limit);
        // totalCount = dto?.TotalCount ?? 0;
        return dto?.Areas ?? new List<Area>();
    }

    /// <summary>
    ///     Uploading the encrypted data to the server
    /// </summary>
    /// <param name="submissionId">Submissions ID</param>
    /// <param name="encryptedAttachments">Encrypted attachments with id and content</param>
    private async Task<bool> UploadAttachmentsAsync(Guid submissionId,
        Dictionary<Guid, string> encryptedAttachments) {
        try {
            foreach (var (id, content) in encryptedAttachments) {
                Logger?.LogInformation("Uploading attachment {Id}", id);
                await SubmissionService.UploadAttachment(submissionId, id, content);
            }

            return true;
        }
        catch (Exception e) {
            Logger?.LogError("Error Uploading attachment {message}", e.Message);
            throw;
        }
    }

    private static void SubmissionSenderGuards(ISendableSubmission submission) {
        // if (submission.DestinationId == null)
        //     throw new ArgumentNullException(nameof(SendableSubmission.DestinationId));
        //if (!Regex.IsMatch(submission.DestinationId, GuidPattern))
        //    throw new ArgumentException("The destination must be a valid GUID");

        if (submission.ServiceName == null)
            throw new ArgumentNullException(nameof(SendableSubmission.ServiceName));

        if (submission.ServiceIdentifier == null)
            throw new ArgumentNullException(nameof(SendableSubmission.ServiceIdentifier));

        if (submission.Data == null)
            throw new ArgumentNullException(nameof(SendableSubmission.Data));

        if (new Regex(LeikaKeyPattern).IsMatch(submission.LeikaKey) == false)
            throw new ArgumentException("The leika key is not valid");

        var destination = await DestinationService.GetDestinationAsync(submission.DestinationId);
        var leikaMatch = destination?.Services.Any(s => s.Identifier == submission.LeikaKey) ??
                         false;
        if (!leikaMatch)
            throw new ArgumentException("The leika key is not valid for the destination");

        if (submission.SchemaUri == null)
            return;

        var schemaMatch = destination?.Services?.Any(s =>
            s.SubmissionSchemas?.Any(schema =>
                schema.SchemaUri == submission.SchemaUri.ToString()) ?? false) ?? false;

        if (!schemaMatch)
            throw new ArgumentException("The schema uri is not valid for the destination");
    }
}

internal static class SubmissionSenderExtension {
    public static void AddServiceType(this Submission submission, string serviceName,
        string serviceIdentifier) {
        if (string.IsNullOrWhiteSpace(serviceIdentifier) || !Regex.IsMatch(serviceIdentifier,
                FitConnectClient.ServiceIdentifierPattern))
            throw new ArgumentException("Invalid leika key");

        submission.ServiceType = new ServiceType {
            Name = serviceName,
            Identifier = serviceIdentifier
        };
    }

    public static void AddData(this Submission submission, string data) {
        try {
            JsonConvert.DeserializeObject(data);
        }
        catch (Exception e) {
            throw new ArgumentException("The data must be valid JSON string", e);
        }

        submission.Data = data;
    }
}
